-- -------------------------------------------------------
-- initialization
-- -------------------------------------------------------

fmBuffCheck = AceLibrary("AceAddon-2.0"):new("AceDB-2.0", "FuBarPlugin-2.0", "AceEvent-2.0", "AceConsole-2.0", "AceComm-2.0")
fmBuffCheck.L = AceLibrary("AceLocale-2.2"):new("fmBuffCheck")
fmBuffCheck.B = AceLibrary("Babble-Spell-2.2")
fmBuffCheck.D = AceLibrary("Dewdrop-2.0")

local self, fmBuffCheck = fmBuffCheck, fmBuffCheck
local L, B, D = fmBuffCheck.L, fmBuffCheck.B, fmBuffCheck.D

fmBuffCheck.commPrefix = "fmbc"
fmBuffCheck.commDist = "GROUP"
fmBuffCheck.player = UnitName("player")
_, fmBuffCheck.pcls = UnitClass("player")
fmBuffCheck.configver = fmBuffCheck_config_version

-- FuBar configuration
fmBuffCheck.hasIcon = true
fmBuffCheck.hideWithoutStandby = true
fmBuffCheck.defaultPosition = "MINIMAP"
fmBuffCheck.defaultMinimapPosition = 200
fmBuffCheck.hasNoText = true
fmBuffCheck.hasNoColor = true
fmBuffCheck.cannotDetachTooltip = true
fmBuffCheck.independentProfile = false
fmBuffCheck.blizzardTooltip = true

fmBuffCheck:RegisterDB("fmBuffCheckDB")
fmBuffCheck:RegisterDefaults("profile", fmBuffCheck_default_config)

fmBuffCheck.version = tonumber(("$Rev: 47057 $"):match("(%d+)"))
fmBuffCheck.date = ("$Date: 2007-08-21 03:41:40 -0400 (Tue, 21 Aug 2007) $"):match("%d%d%d%d%-%d%d%-%d%d")

-- default variables and values
fmBuffCheck.repver = {}
fmBuffCheck.repbuf = {}
fmBuffCheck.recver = 0
fmBuffCheck.recbuf = 0
fmBuffCheck.cntver = 0
fmBuffCheck.cntbuf = 0
fmBuffCheck.lastbuffresults = {}
	
-- menu structure
fmBuffCheck.options = {
	type = 'group',
	handler = fmBuffCheck,
	args = {
		debugMode = {
			order = 10,
			name = L["Debugging Mode"],
			desc = L["If enabled, additional messages are printed out to help debugging."],
			type = 'toggle',
			get = function() return self.db.profile.debugging end,
			set = function(v) self.db.profile.debugging = v end,
		},
		nagMode = {
			order = 20,
			name = L["Nagging Mode"],
			desc = L["If enabled, every person NOT having all buffs is being reminded by whisper."],
			type = 'toggle',
			get = function() return self.db.profile.nagging end,
			set = function(v) self.db.profile.nagging = v end,
		},
		reqTimeout = {
			order = 30,
			name = L["Request Timeout"],
			desc = L["The timeout of requests sent out (seconds to wait for answers)."],
			min = 1,
			max = 20,
			step = 1,
			type = 'range',
			get = function() return self.db.profile.requesttimeout end,
			set = function(v) self.db.profile.requesttimeout = v end,
		},
		bufTimeout = {
			order = 35,
			name = L["Buff Timeout"],
			desc = L["The time a buff has to be running to be considered valid (minutes)."],
			min = 0,
			max = 30,
			step = 1,
			type = 'range',
			get = function() return self.db.profile.bufftimeout end,
			set = function(v) self.db.profile.bufftimeout = v end,
		},
		shwLogframe = {
			order = 36,
			name = L["Show Log Frame"],
			desc = L["Show the Log frame with extended information."],
			type = 'toggle',
			get = function() return self.db.profile.showlogframe end,
			set = function(v)
					   self.db.profile.showlogframe = v
					   if (v == false) then 
							fmBuffCheck:LogFrameHide()
					   end
				  end,
		},
		shwResframe = {
			order = 37,
			name = L["Show Result Frame"],
			desc = L["Show the Result frame with concise information."],
			type = 'toggle',
			get = function() return self.db.profile.showresframe end,
			set = function(v)
						self.db.profile.showresframe = v
						fmBuffCheck:ResFrameShow()
				  end,
		},
		spacer1 = {
			order = 40,
			type = 'header',
		},
		buffProfileSel = {
			order = 50,
			name = L["Active Buff Profile:"].." "..L["none"],
			desc = L["Active Buff Profile:"].." "..L["none"],
			type = 'group',
			hidden = true,
			args = {},
		},
		baseProfileSel = {
			order = 60,
			name = L["Active Base Profile:"].." "..L["none"],
			desc = L["Active Base Profile:"].." "..L["none"],
			type = 'group',
			hidden = true,
			args = {},
		},
		spacer2 = {
			order = 70,
			type = 'header',
		},
		buffProfiles = {
			order = 80,
			name = L["Edit Buff Profiles"],
			desc = L["List of available buff profiles."],
			type = 'group',
			hidden = true,
			args = {},
		},
		buffProfileAdd = {
			order = 90,
			name = L["Add New Buff Profile"],
			desc = L["Add a new buff profile."],
			type = 'text',
			usage = "<ProfileName>",
			set = "BuffProfileAdd",
			get = function() return "" end,
		},
		buffProfileDel = {
			order = 100,
			name = L["Delete Buff Profile"],
			desc = L["Delete a buff profile."],
			type = 'group',
			hidden = true,
			args = {},
		},
		buffProfileRen = {
			order = 110,
			name = L["Rename Buff Profile"],
			desc = L["Rename a buff profile."],
			type = 'group',
			hidden = true,
			args = {},
		},
		buffProfileCopy = {
			order = 120,
			name = L["Copy Buff Profile"],
			desc = L["Copy a buff profile."],
			type = 'group',
			hidden = true,
			args = {},
		},
		spacer3 = {
			order = 130,
			type = 'header',
		},
		baseProfiles = {
			order = 140,
			name = L["Edit Base Profiles"],
			desc = L["List of available base profiles."],
			type = 'group',
			hidden = true,
			args = {},
		},
		baseProfileAdd = {
			order = 150,
			name = L["Add New Base Profile"],
			desc = L["Add a new base profile."],
			type = 'text',
			usage = "<ProfileName>",
			set = "BaseProfileAdd",
			get = function() return "" end,
		},
		baseProfileDel = {
			order = 160,
			name = L["Delete Base Profile"],
			desc = L["Delete a base profile."],
			type = 'group',
			hidden = true,
			args = {},
		},
		baseProfileRen = {
			order = 170,
			name = L["Rename Base Profile"],
			desc = L["Rename a base profile."],
			type = 'group',
			hidden = true,
			args = {},
		},
		baseProfileCopy = {
			order = 180,
			name = L["Copy Base Profile"],
			desc = L["Copy a base profile."],
			type = 'group',
			hidden = true,
			args = {},
		},
	}
}

fmBuffCheck.OnMenuRequest = fmBuffCheck.options

-- -------------------------------------------------------
-- functions
-- -------------------------------------------------------

function fmBuffCheck:OnInitialize()
	-- create frames
	self:CreateLogFrame()
	self:CheckConfig()
	self:SetCommPrefix(self.commPrefix)
	self:RegisterComm(self.commPrefix, self.commDist, "OnCommReceive")
	self:UpdateTooltip()
	self:UpdateMenuStructure()
	self:ResFrameShow()
	self:ResFrameWrite(L["Waiting"], L["for first"], L["update"])
end

function fmBuffCheck:OnEnable()
	self:RegisterChatCommand('/fmBuffCheck', '/fmbc', {
		type = 'execute',
		func = function()
			self:CmdI2buf()
		end
	})
end

function fmBuffCheck:Send(cmd, res)
	self:SendCommMessage(self.commDist, cmd, res)
	self:OnCommReceive(self.commPrefix, self.player, self.commDist, cmd, res)
end

function fmBuffCheck:OnCommReceive(prefix, sender, distribution, cmd, res)
	
	if     cmd == "Q2VER" then self:CmdQ2ver(sender, cmd, res) -- received query for version info
	elseif cmd == "Q2BUF" then self:CmdQ2buf(sender, cmd, res) -- received query for buff info
	elseif cmd == "R2VER" then self:CmdR2ver(sender, cmd, res) -- received report of version info
	elseif cmd == "R2BUF" then self:CmdR2buf(sender, cmd, res) -- received report of buff info
	else                       self:CmdUnkwn(sender, cmd, res) -- received unknown command
	end
	
end

function fmBuffCheck:CmdI2ver(override)
	-- request an update of all versions from group members
	if(self:IsAuthorized() or override) then
		self:DebugPrint(L["Requesting version status of all party members..."])
		self:LogFrameClear()
		self.repver = {}
		self.cntver = 0
		self.recver = 0
		self:LogFrameShow()
		self:LogFrameAdd((L["Version request in progress, %s seconds remaining, got %s of %s."]):format(self.db.profile.requesttimeout, self.recver, self:GroupSize()))
		self:Send("Q2VER", false)
		self:ScheduleEvent("fmBuffCheck-DisplayVersion", self.DisplayVersions, 1)
	end
end

function fmBuffCheck:CmdI2buf(override)
	-- request an update of all buffs from group members
	if(self:IsAuthorized() or override) then
		self:DebugPrint(L["Requesting buff status of all party members..."])
		self:LogFrameClear()
		--self:ResFrameClear()
		self.repbuf = {}
		self.cntbuf = 0
		self.recbuf = 0
		self.lastbuffresults = {}
		self:LogFrameShow()
		self:ResFrameShow()
		self:LogFrameAdd((L["Buff request in progress, %s seconds remaining, got %s of %s."]):format(self.db.profile.requesttimeout, self.recbuf, self:GroupSize()))
		self:Send("Q2BUF", false)
		self:ScheduleEvent("fmBuffCheck-DisplayBuff", self.DisplayBuffs, 1)
	end
end

function fmBuffCheck:CmdQ2ver(sender, cmd, res)
	 -- received query for version info
	self:DebugPrint((L["Version list request received from %s."]):format(sender))
	self:Send("R2VER", ("%s (%s)"):format(self.version, self.date))
end

function fmBuffCheck:CmdQ2buf(sender, cmd, res)
	-- received query for buff info
	self:DebugPrint((L["Buff list request received from %s."]):format(sender))
	self:GetAndSendBuffs()
end

function fmBuffCheck:CmdR2ver(sender, cmd, res)
	-- received report of version info
	self:DebugPrint((L["Version list result received from %s."]):format(sender))
	self.repver[sender] = res
	self.recver = self.recver + 1
end

function fmBuffCheck:CmdR2buf(sender, cmd, res)
	-- received report of buff info
	self:DebugPrint((L["Buff list result received from %s."]):format(sender))
	self.repbuf[sender] = res
	self.recbuf = self.recbuf + 1
end

function fmBuffCheck:CmdUnkwn(sender, cmd, res)
	-- received unknown command
	self:DebugPrint((L["Unknown QUERY request received from %s: %s."]):format(sender, cmd))
end

function fmBuffCheck:GetAndSendBuffs()
	local num, k, v
	local buffs = {}

	self:DebugPrint("Starting local Buff processing...")
	
	for num = 1, 40 do
		local id = GetPlayerBuff(num, "HELPFUL")
		if id > 0 then
			local name = GetPlayerBuffName(id)
			local time = GetPlayerBuffTimeLeft(id)
			local appl = GetPlayerBuffApplications(id)
			local ttip = self:GetBuffTooltipById(id)
			name = self:ReverseTranslateSafe(name) -- always try to get it in English!
			
			buffs[#buffs + 1] = {
				["type"] = "buf",
				["name"] = name,
				["time"] = time,
				["appl"] = appl,
				["ttip"] = ttip
			}
		end
	end
	
	local hasMainHandEnchant, mainHandExpiration, mainHandCharges, hasOffHandEnchant, offHandExpiration, offHandCharges = GetWeaponEnchantInfo(); 
	if hasMainHandEnchant then
		buffs[#buffs + 1] = {
			["type"] = "mhw",
			["name"] = self:WeaponParse("MainHandSlot"),
			["time"] = mainHandExpiration / 1000,
			["appl"] = mainHandCharges,
			["ttip"] = self:GetItemTooltip("MainHandSlot")
		}
	end
	if hasOffHandEnchant then
		buffs[#buffs + 1] = {
			["type"] = "ohw",
			["name"] = self:WeaponParse("SecondaryHandSlot"),
			["time"] = offHandExpiration / 1000,
			["appl"] = offHandCharges,
			["ttip"] = self:GetItemTooltip("SecondaryHandSlot")
		}
	end
	
	if self.db.profile.debugging then 
		for k,v in pairs(buffs) do
			self:LogFrameAdd(("%s: %s (%s)\n%s\n---------"):format(v["type"], v["name"], v["time"], v["ttip"]))
		end
	end
	
	self:Send("R2BUF", buffs)
end

function fmBuffCheck:GetBuffTooltipById(id)
	local i
	local ret = ""
	
	fmBuffCheckScanningTooltip:ClearLines()
	fmBuffCheckScanningTooltip:SetPlayerBuff(id)
	
	for i = 1, fmBuffCheckScanningTooltip:NumLines() do
		local txt = getglobal("fmBuffCheckScanningTooltipTextLeft" .. i):GetText()
		ret = ret .. txt .. "\n"
	end
	return ret
end

function fmBuffCheck:GetItemTooltip(slot)
	fmBuffCheckScanningTooltip:ClearLines()
	local slotId, _ = GetInventorySlotInfo(slot)
	fmBuffCheckScanningTooltip:SetInventoryItem(self.player, slotId)
	local ret = ""
	
	for i = 1, fmBuffCheckScanningTooltip:NumLines() do
		local txt = getglobal("fmBuffCheckScanningTooltipTextLeft" .. i):GetText()
		ret = ret .. txt .. "\n"
    end
    return ret
end

function fmBuffCheck:StringContains(haystack, needle)
	return (string.find(haystack, needle, 1, true) or string.find(haystack, self:TranslateSafe(needle), 1, true))
end

function fmBuffCheck:WeaponParse(slot)
	fmBuffCheckScanningTooltip:ClearLines()
	local slotId, _ = GetInventorySlotInfo(slot)
	fmBuffCheckScanningTooltip:SetInventoryItem(self.player, slotId)
	return getglobal("fmBuffCheckScanningTooltipTextLeft1"):GetText()
end

function fmBuffCheck:FormatTime(time)
	if time > 60 then
		return math.floor(time / 60) .. ":" .. ("%02d"):format(math.floor(time) % 60) .. "m"
	end
	return math.floor(time) .. "s"
end

function fmBuffCheck:FormatTimeShort(time)
	if time == 0 then
		return "\226\136\158"
	end
	if time > 60 then
		return math.floor(time / 60) .. "m"
	end
	return math.floor(time) .. "s"
end

function fmBuffCheck:PP(obj, style)
	local name = obj["name"]
	local timeleft = obj["time"]

	if (style == 1) then
		if (name == nil or name == "") then
			return string.format("|cffff0000%s|r", L["none"]) 
		end
	
		return string.format("|cff00ff00%s (%s)|r", name, self:FormatTime(timeleft)) 
	else
		if (name == nil or name == "") then
			return string.format("|cffff0000%s|r", L["none"]) 
		end
	
		return string.format("|cff00ff00(%s)|r", self:FormatTime(timeleft)) 
	end
end

function fmBuffCheck:DebugPrint(msg)
	if self.db.profile.debugging then
		self.Print(msg)
	end
end

function fmBuffCheck:OnTooltipUpdate()
	GameTooltip:AddLine("fmBuffCheck")
	GameTooltip:AddLine(" ")
	GameTooltip:AddDoubleLine(L["Nagging:"], (self.db.profile.nagging and YES or NO), 1, 1, 0, 1, 1, 1)
	GameTooltip:AddDoubleLine(L["Debugging:"], (self.db.profile.debugging and YES or NO), 1, 1, 0, 1, 1, 1)
	GameTooltip:AddDoubleLine(L["Active Buff Profile:"], self:GetActiveBuffProfile(), 1, 1, 0, 1, 1, 1)
	GameTooltip:AddDoubleLine(L["Active Base Profile:"], self:GetActiveBaseProfile(), 1, 1, 0, 1, 1, 1)
	GameTooltip:AddDoubleLine(L["Request Timeout:"], self.db.profile.requesttimeout .. "s", 1, 1, 0, 1, 1, 1)
	GameTooltip:AddDoubleLine(L["Buff Timeout:"], self.db.profile.bufftimeout .. "m", 1, 1, 0, 1, 1, 1)
	GameTooltip:AddLine(" ")
	GameTooltip:AddLine("|cffffff00" .. L["Click|r to initiate buff check."], 0.2, 1, 0.2)
	GameTooltip:AddLine("|cffffff00" .. L["Shift-Click|r to initiate version check."], 0.2, 1, 0.2)
end

function fmBuffCheck:OnClick()
	local shift, alt, ctrl = IsShiftKeyDown(), IsAltKeyDown(), IsControlKeyDown()
	if not shift and not alt then
		self:CmdI2buf(ctrl)
	end
	if shift and not alt then
		self:CmdI2ver(ctrl)
	end
end

function fmBuffCheck:IsGrouped()
	local inRaid = (GetNumRaidMembers() > 0)
	local inGroup = (GetNumPartyMembers() > 0)
	return (inRaid or inGroup)
end

function fmBuffCheck:GroupSize()
	if (GetNumRaidMembers() > 0) then
		return GetNumRaidMembers()
	end
	if (GetNumPartyMembers() > 0) then
		return GetNumPartyMembers() + 1
	end
	return 1
end

function fmBuffCheck:IsAuthorized()
	return (not self:IsGrouped() or IsPartyLeader() or IsRaidLeader() or IsRaidOfficer() or self.player == "Nominedomini")
end

function fmBuffCheck:BuffProfileAdd(value)
	local cnt = #self.db.profile.buffprofile
	
	if (cnt == nil or cnt == 0) then
		self.db.profile.buffprofile = {}
		cnt = 0
	end
	
	cnt = cnt + 1
	
	self.db.profile.buffprofile[cnt] = {}
	self.db.profile.buffprofile[cnt].name = value
	self.db.profile.buffprofile[cnt].profile = fmBuffCheck_default_buffprofile
	self:UpdateMenuStructure()
end

function fmBuffCheck:BaseProfileAdd(value)
	local cnt = #self.db.profile.baseprofile
	
	if (cnt == nil or cnt == 0) then
		self.db.profile.baseprofile = {}
		cnt = 0
	end
	
	cnt = cnt + 1
	
	self.db.profile.baseprofile[cnt] = {}
	self.db.profile.baseprofile[cnt].name = value
	self.db.profile.baseprofile[cnt].profile = fmBuffCheck_default_baseprofile
	self:UpdateMenuStructure()
end

function fmBuffCheck:BuffProfileCopy(i, value)
	local cnt = #self.db.profile.buffprofile
	self.db.profile.buffprofile[cnt+1].name = value
	self.db.profile.buffprofile[cnt+1].profile = self.db.profile.buffprofile[i].profile
	self:UpdateMenuStructure()
end

function fmBuffCheck:BaseProfileCopy(i, value)
	local cnt = #self.db.profile.baseprofile
	self.db.profile.baseprofile[cnt+1].name = value
	self.db.profile.baseprofile[cnt+1].profile = self.db.profile.baseprofile[i].profile
	self:UpdateMenuStructure()
end

function fmBuffCheck:BuffProfileRen(i, value)
	self.db.profile.buffprofile[i].name = value
	self:UpdateMenuStructure()
end

function fmBuffCheck:BaseProfileRen(i, value)
	self.db.profile.baseprofile[i].name = value
	self:UpdateMenuStructure()
end

function fmBuffCheck:BuffProfileSel(value)
	self.db.profile.activebuffprofile = value
	self:UpdateMenuStructure()
end

function fmBuffCheck:BaseProfileSel(value)
	self.db.profile.activebaseprofile = value
	self:UpdateMenuStructure()
end

function fmBuffCheck:BuffProfileDel(value)
	local cnt = #self.db.profile.buffprofile
	
	if (cnt == nil or cnt == 0 or value > cnt) then
		self:UpdateMenuStructure()
		return
	end
	
	if (value < cnt) then
		-- move profiles down (boring)
		for i = value + 1, cnt do
			self.db.profile.buffprofile[i-1] = self.db.profile.buffprofile[i]
		end
	end
	self.db.profile.buffprofile[cnt] = nil
	self:UpdateMenuStructure()
end

function fmBuffCheck:BaseProfileDel(value)
	local cnt = #self.db.profile.baseprofile
	
	if (cnt == nil or cnt == 0 or value > cnt) then
		self:UpdateMenuStructure()
		return
	end
	
	if (value < cnt) then
		-- move profiles down (boring)
		for i = value + 1, cnt do
			self.db.profile.baseprofile[i-1] = self.db.profile.baseprofile[i]
		end
	end
	self.db.profile.baseprofile[cnt] = nil
	self:UpdateMenuStructure()
end


function fmBuffCheck:UpdateMenuStructure()
	-- clear lists
	self.options.args.buffProfiles.args = {}
	self.options.args.buffProfileDel.args = {}
	self.options.args.buffProfileRen.args = {}
	self.options.args.buffProfileSel.args = {}
	self.options.args.buffProfileCopy.args = {}

	self.options.args.baseProfiles.args = {}
	self.options.args.baseProfileDel.args = {}
	self.options.args.baseProfileRen.args = {}
	self.options.args.baseProfileSel.args = {}
	self.options.args.baseProfileCopy.args = {}
	
	-- hide lists
	self.options.args.buffProfiles.hidden = true
	self.options.args.buffProfileDel.hidden = true
	self.options.args.buffProfileRen.hidden = true
	self.options.args.buffProfileSel.hidden = true
	self.options.args.buffProfileCopy.hidden = true
	
	self.options.args.baseProfiles.hidden = true
	self.options.args.baseProfileDel.hidden = true
	self.options.args.baseProfileRen.hidden = true
	self.options.args.baseProfileSel.hidden = true
	self.options.args.baseProfileCopy.hidden = true
	
	-- update stuff
	self.options.args.buffProfileSel.name = L["Active Buff Profile:"] .. " " .. self:GetActiveBuffProfile()
	self.options.args.buffProfileSel.desc = L["Active Buff Profile:"] .. " " .. self:GetActiveBuffProfile()

	self.options.args.baseProfileSel.name = L["Active Base Profile:"] .. " " .. self:GetActiveBaseProfile()
	self.options.args.baseProfileSel.desc = L["Active Base Profile:"] .. " " .. self:GetActiveBaseProfile()
		
	if (#self.db.profile.buffprofile) then
		local i		
		for i = 1, #self.db.profile.buffprofile do
			self.options.args.buffProfiles.args[i] = {
				name = self.db.profile.buffprofile[i].name or "?Oo?",
				desc = self.db.profile.buffprofile[i].name or "?Oo?",
				type = "group",
				args = self:GetBuffProfile(i),
			}
			self.options.args.buffProfileDel.args[i] = {
				name = self.db.profile.buffprofile[i].name or "?Oo?",
				desc = self.db.profile.buffprofile[i].name or "?Oo?",
				type = "execute",
				func = "BuffProfileDel",
				passValue = i,
			}
			self.options.args.buffProfileRen.args[i] = {
				name = self.db.profile.buffprofile[i].name or "?Oo?",
				desc = self.db.profile.buffprofile[i].name or "?Oo?",
				type = "text",
				usage = "<ProfileName>",
				set = "BuffProfileRen",
				get = function() return self.db.profile.buffprofile[i].name end,
				passValue = i,
			}		
			self.options.args.buffProfileSel.args[i] = {
				name = self.db.profile.buffprofile[i].name or "?Oo?",
				desc = self.db.profile.buffprofile[i].name or "?Oo?",
				type = "execute",
				func = "BuffProfileSel",
				passValue = i,
			}
			self.options.args.buffProfileCopy.args[i] = {
				name = self.db.profile.buffprofile[i].name or "?Oo?",
				desc = self.db.profile.buffprofile[i].name or "?Oo?",
				type = "text",
				usage = "<ProfileName>",
				set = "BuffProfileCopy",
				get = function() return self.db.profile.buffprofile[i].name end,
				passValue = i,
			}
		end
		
		self.options.args.buffProfiles.hidden = false
		self.options.args.buffProfileDel.hidden = false
		self.options.args.buffProfileRen.hidden = false
		self.options.args.buffProfileSel.hidden = false
		self.options.args.buffProfileCopy.hidden = false
	end

	if (#self.db.profile.baseprofile) then
		local i		
		for i = 1, #self.db.profile.baseprofile do
			self.options.args.baseProfiles.args[i] = {
				name = self.db.profile.baseprofile[i].name or "?Oo?",
				desc = self.db.profile.baseprofile[i].name or "?Oo?",
				type = "group",
				args = self:GetBaseProfile(i),
			}
			self.options.args.baseProfileDel.args[i] = {
				name = self.db.profile.baseprofile[i].name or "?Oo?",
				desc = self.db.profile.baseprofile[i].name or "?Oo?",
				type = "execute",
				func = "BaseProfileDel",
				passValue = i,
			}
			self.options.args.baseProfileRen.args[i] = {
				name = self.db.profile.baseprofile[i].name or "?Oo?",
				desc = self.db.profile.baseprofile[i].name or "?Oo?",
				type = "text",
				usage = "<ProfileName>",
				set = "BaseProfileRen",
				get = function() return self.db.profile.baseprofile[i].name end,
				passValue = i,
			}		
			self.options.args.baseProfileSel.args[i] = {
				name = self.db.profile.baseprofile[i].name or "?Oo?",
				desc = self.db.profile.baseprofile[i].name or "?Oo?",
				type = "execute",
				func = "BaseProfileSel",
				passValue = i,
			}
			self.options.args.baseProfileCopy.args[i] = {
				name = self.db.profile.baseprofile[i].name or "?Oo?",
				desc = self.db.profile.baseprofile[i].name or "?Oo?",
				type = "text",
				usage = "<ProfileName>",
				set = "BaseProfileCopy",
				get = function() return self.db.profile.baseprofile[i].name end,
				passValue = i,
			}
		end
		
		self.options.args.baseProfiles.hidden = false
		self.options.args.baseProfileDel.hidden = false
		self.options.args.baseProfileRen.hidden = false
		self.options.args.baseProfileSel.hidden = false
		self.options.args.baseProfileCopy.hidden = false
	end

end

function fmBuffCheck:GetBuffProfile(i)
	local t = {}
	local k1, v1
	
	for k1,v1 in pairs(self.db.profile.baseprofile[self.db.profile.activebaseprofile]["profile"]) do
		t[k1..i] = {
			name = self:TranslateSafe(k1),
			desc = self:TranslateSafe(k1),
			type = "toggle",
			get = function() return self.db.profile.buffprofile[i]["profile"][k1] end,
			set = function(v) self.db.profile.buffprofile[i]["profile"][k1] = v end,
		}
	end
	return t
end

function fmBuffCheck:GetBaseProfile(i)
	local t = {}
	local k1, v1
	for k1,v1 in pairs(self.db.profile.baseprofile[i]["profile"]) do
		t[#t+1] = {
			name = self:TranslateSafe(k1),
			desc = self:TranslateSafe(k1),
			type = "group",
			args = self:GetBaseProfileSub(i, k1)
		}
	end
	return t
end

function fmBuffCheck:GetBaseProfileSub(id, na)
	local t = {}
	local k1, v1

	for k1,v1 in pairs(self.db.profile.baseprofile[id]["profile"][na]["sources"]) do
		t[#t+1] = {
			name = ("%s (%1.1f)"):format(self:TranslateSafe(k1), v1.Weight),
			desc = ("%s (%1.1f)"):format(self:TranslateSafe(k1), v1.Weight),
			type = "group",
			args = self:GetBaseProfileSubSub(id, na, k1)
		}
	end
	return t
end

function fmBuffCheck:GetBaseProfileSubSub(id, na, so)
	local t = {}
	local k1, v1

	for k1,v1 in pairs(self.db.profile.baseprofile[id]["profile"][na]["sources"][so]) do
		t[#t+1] = {
			name = ("%s (%1.1f)"):format(self:TranslateSafe(k1), v1),
			desc = ("%s (%1.1f)"):format(self:TranslateSafe(k1), v1),
			min = 0,
			max = 1,
			step = 0.1,
			type = 'range',
			get = function() return self.db.profile.baseprofile[id]["profile"][na]["sources"][so][k1] end,
			set = function(v)
						self.db.profile.baseprofile[id]["profile"][na]["sources"][so][k1] = v
						self:UpdateMenuStructure()
				  end,
		}
	end
	return t
end

function fmBuffCheck:GetActiveBuffProfile()
	local cnt = #self.db.profile.buffprofile
	local act = self.db.profile.activebuffprofile
		
	if (cnt == nil or cnt == 0) then
		self.db.profile.activebuffprofile = nil
		return ""
	end
	
	if (act == nil or act == 0) then
		self.db.profile.activebuffprofile = nil
		return ""	
	end
	
	if (act > cnt) then
		self.db.profile.activebuffprofile = nil
		return ""
	end
	
	return self.db.profile.buffprofile[act].name
end

function fmBuffCheck:GetActiveBaseProfile()
	local cnt = #self.db.profile.baseprofile
	local act = self.db.profile.activebaseprofile
		
	if (cnt == nil or cnt == 0) then
		self.db.profile.activebaseprofile = nil
		return ""
	end
	
	if (act == nil or act == 0) then
		self.db.profile.activebaseprofile = nil
		return ""	
	end
	
	if (act > cnt) then
		self.db.profile.activebaseprofile = nil
		return ""
	end
	
	return self.db.profile.baseprofile[act].name
end

function fmBuffCheck:ReverseTranslateSafe(txt)
	return B:HasReverseTranslation(txt) and B:GetReverseTranslation(txt) or L:HasReverseTranslation(txt) and L:GetReverseTranslation(txt) or txt
end

function fmBuffCheck:TranslateSafe(txt)
	return B:HasTranslation(txt) and B:GetTranslation(txt) or L:HasTranslation(txt) and L:GetTranslation(txt) or txt
end

function fmBuffCheck:DisplayBuffs()
	local fmBuffCheck, self = fmBuffCheck, fmBuffCheck
	local grpsiz = self:GroupSize()
	
	self.cntbuf = self.cntbuf + 1
	
	-- still counting down?
	if (self.cntbuf < self.db.profile.requesttimeout and self.recbuf < grpsiz) then
		self:LogFrameClear()
		self:LogFrameAdd((L["Buff request in progress, %s seconds remaining, got %s of %s."]):format(self.db.profile.requesttimeout - self.cntbuf, self.recbuf, grpsiz))
		self:ScheduleEvent("fmBuffCheck-DisplayBuff", self.DisplayBuffs, 1)
	-- or ready to display (after timeout)?
	else
		local i, k, v, k1, v1, name, pcls
		local classFact = self:GetClassFact()
		local players = {}
		
		self:LogFrameClear()
		self:LogFrameAdd((L["Buff request completed, got %s of %s answers:"]):format(self.recbuf, grpsiz))
		
		for k, v in pairs(classFact) do
			self:DebugPrint(("%s: %s"):format(k, v))
		end
		
		-- are we a raid?
		if (GetNumRaidMembers() > 0) then
			-- go through each raid member and display the results
			for i = 1, GetNumRaidMembers() do
				local name, _, _, _, pcls, _, _, _, _, _, _ = GetRaidRosterInfo(i);
				players[name] = {}
				players[name]["name"]   = name
				players[name]["source"] = self.repbuf[name]
				players[name]["pcls"]   = self:ReverseTranslateSafe(self:Ucfirst(pcls))
				players[name]["fact"]   = classFact[players[name]["pcls"]]
				players[name]["values"] = {}
			end 
		-- are we a group (or alone)?
		else
			-- are we a group?
			if (GetNumPartyMembers() > 0) then
				for i = 1, GetNumPartyMembers() do
					local name, _ = UnitName("party" .. i)
					local _, pcls = UnitClass("party" .. i);
					players[name] = {}
					players[name]["name"]   = name
					players[name]["source"] = self.repbuf[name]
					players[name]["pcls"]   = self:ReverseTranslateSafe(self:Ucfirst(pcls))
					players[name]["fact"]   = classFact[players[name]["pcls"]]
					players[name]["values"] = {}
				end
			end
			-- always process ourself
			name = self.player
			pcls = self.pcls
			players[name] = {}
			players[name]["name"]   = name
			players[name]["source"] = self.repbuf[name]
			players[name]["pcls"]   = self:ReverseTranslateSafe(self:Ucfirst(pcls))
			players[name]["fact"]   = classFact[players[name]["pcls"]]
			players[name]["values"] = {}
		end
		
		-- now that we have all numbers in place, lets do some analysis...

		local t =
		{
			{
				t = 10,
				b = 0,
				w = 100,
				a = 0,
				n = "",
				s = "",
			}, --[1]
			{
				t = self.db.profile.bufftimeout * 60,
				b = 0,
				w = 100,
				a = 0,
				n = "",
				s = "",
			}, --[2]
		}

		-- 1. go through all requested times (could be multiple)
		for k1, v1 in pairs(t) do
			local s = 0
			local c = 0
			-- 2. determine all buff values and the best/average/worst player
			for k, v in pairs(players) do
				local pv, pvw
				players[k].values[k1] = {}
				players[k].values[k1].t0 = v1.t
				pv = self:GetBuffValue(players[k]["name"], players[k]["pcls"], v1.t)
				players[k].values[k1].value = pv
				pvw = pv * players[k]["fact"]
				
				if (pvw <= v1.w) then
					v1.w = pvw
					v1.n = players[k]["name"]
				end
				if (pvw >= v1.b) then
					v1.b = pvw 
				end
				s = s + pvw
				c = c + 1
			end	
			v1.a = s / c
			self:LogFrameAdd(("%s: %3.0f%% (%s)/ %3.0f%% / %3.0f%%"):format(self:FormatTime(v1.t), v1.w, v1.n, v1.a, v1.b))
			v1.s = ("%s: %3.0f / %3.0f / %3.0f"):format(self:FormatTimeShort(v1.t), v1.w, v1.a, v1.b)	
		end
		
		self:ResFrameWrite(L["Results:"], t[1].s, t[2].s)
		
		self.lastbuffresults.p = players
		self.lastbuffresults.t = t
		
		D:Register(fmBuffCheckResFrame, 'children', function() D:FeedAceOptionsTable(self:GetResFrameMenu()) end)
	end
end

function fmBuffCheck:DisplayVersions()
	local i, k, v, name, grpsiz
	local fmBuffCheck, self = fmBuffCheck, fmBuffCheck
	
	self.cntver = self.cntver + 1
	grpsiz = self:GroupSize()
	
	-- still counting down?
	if (self.cntver < self.db.profile.requesttimeout and self.recver < grpsiz) then
		self:LogFrameClear()
		self:LogFrameAdd((L["Version request in progress, %s seconds remaining, got %s of %s."]):format(self.db.profile.requesttimeout - self.cntver, self.recver, grpsiz))
		self:ScheduleEvent("fmBuffCheck-DisplayVersion", self.DisplayVersions, 1)
	-- or ready to display (after timeout)?
	else
		self:LogFrameClear()
		
		self:LogFrameAdd((L["Version request completed, got %s of %s answers:"]):format(self.recver, grpsiz))
		-- are we a raid?
		if (GetNumRaidMembers() > 0) then
			-- go through each raid member and display the results
			for i = 1, GetNumRaidMembers() do
				name, _, _, _, _, _, _, _, _, _, _ = GetRaidRosterInfo(i);
				self:ProcessVersion(name)
			end 
		-- are we a group?
		elseif (GetNumPartyMembers() > 0) then
			self:ProcessVersion(self.player)
			for i = 1, GetNumPartyMembers() do
				name, _ = UnitName("party" .. i)
				self:ProcessVersion(name)
			end
		-- are we alone?
		else
			-- just display the results (should be only us)
			self:ProcessVersion(self.player)
		end
	end
end

function fmBuffCheck:ProcessVersion(sender)
	local fmBuffCheck, self = fmBuffCheck, fmBuffCheck
	if (self.repver[sender]) then
		self:LogFrameAdd(("|cff00ff00%s: %s|r"):format(sender, self.repver[sender]))
	else
		self:LogFrameAdd(("|cffff0000%s: %s|r"):format(sender, L["none"]))
	end
end

function fmBuffCheck:CheckConfig()
	if ((not self.db.profile) or self.db.profile.configver == nil or self.db.profile.configver < self.configver) then
		--any valid data?
		if (not self.db.profile) then
			self:DebugPrint(L["Invalid Configuration found. Reset to defaults."])
			self.db.profile = fmBuffCheck_default_config
		end
		--try update first
		self:DebugPrint(L["Outdated configuration found, trying to convert."])
		if self:ConfigUpdate(self.db.profile.configver, self.configver) then
			-- if it worked, be happy and rejoice
			self:DebugPrint(L["Configuration converted to current version. Please check."])
		else
			-- if it failed, reset config to defaults... (sad but necessary)
			self:DebugPrint(L["Configuration could not be converted to current version. Reset to defaults."])	
			self.db.profile = fmBuffCheck_default_config
		end
	end
end

function fmBuffCheck:ConfigUpdate(curver, actver)
	if ((curver == nil or curver == 1) and actver == 2) then -- convert from 1 to 2
		self.db.profile.configver = actver
		self.db.profile.baseprofile = {
			{
				["name"] = "Default",
				["profile"] = fmBuffCheck_default_baseprofile,
			} -- 1
		}

		--	buffprofile = {
		--		{
		--			["name"] = "Default",
		--			["profile"] = fmBuffCheck_default_buffprofile,
		--		},
		--		{
		--			...
		--		}
		--	}
		return false -- change to true once the conversion is successful 
	else
		return false
	end
end

function fmBuffCheck:GetClassFact()
	local k1, v1, k2, v2, k3, v3
	
	local classMax = {
		["Warlock"] = 0,
		["Warrior"] = 0,
		["Hunter"]  = 0,
		["Mage"]    = 0,
		["Priest"]  = 0,
		["Druid"]   = 0,
		["Paladin"] = 0,
		["Shaman"]  = 0,
		["Rogue"]   = 0,
	}

	local classFact = {
		["Warlock"] = 1,
		["Warrior"] = 1,
		["Hunter"]  = 1,
		["Mage"]    = 1,
		["Priest"]  = 1,
		["Druid"]   = 1,
		["Paladin"] = 1,
		["Shaman"]  = 1,
		["Rogue"]   = 1,
	}
	
	if (self:GetActiveBuffProfile() == "") then
		self:DebugPrint("No active buff profile")
		return nil
	end

	if (self:GetActiveBaseProfile() == "") then
		self:DebugPrint("No active base profile")
		return nil
	end
	
	for k1,v1 in pairs(self.db.profile.buffprofile[self.db.profile.activebuffprofile]["profile"]) do
		if (v1) then
			local subMax = {
				["Warlock"] = 0,
				["Warrior"] = 0,
				["Hunter"]  = 0,
				["Mage"]    = 0,
				["Priest"]  = 0,
				["Druid"]   = 0,
				["Paladin"] = 0,
				["Shaman"]  = 0,
				["Rogue"] = 0,
			}

			local source = self.db.profile.baseprofile[self.db.profile.activebaseprofile]["profile"][k1]["sources"]
			for k2,v2 in pairs(source) do
				local weight = v2["Weight"]
				for k3, v3 in pairs(subMax) do
					local oldVal = v3
					local newVal = v2[k3]*weight
					subMax[k3] = (newVal > oldVal) and newVal or oldVal
				end
			end
			
			for k3, v3 in pairs(classMax) do
				classMax[k3] = classMax[k3] + subMax[k3]
			end
		end
	end
	
	for k3, v3 in pairs(classMax) do
		classFact[k3] = 100/v3
	end

	return classFact
end

function fmBuffCheck:GetBuffValue(sender, pclso, secs)
	local buffval = 0
	
	if (not self.repbuf[sender]) then
		if self.db.profile.debugging then
			self:LogFrameAdd(("|cffff0000%s: %s|r"):format(sender, L["none"]))
		end
	else
		if (self:GetActiveBuffProfile() > "" and self:GetActiveBaseProfile() > "") then
			local requiredbuffs = {}
			local k1, v1, k2, v2, k3, v3
			local res = self.repbuf[sender]
			local actbuff = self.db.profile.buffprofile[self.db.profile.activebuffprofile]
			local actbase = self.db.profile.baseprofile[self.db.profile.activebaseprofile]
			local pcls    = self:ReverseTranslateSafe(self:Ucfirst(pclso))
			
			if (type(res) ~= "table") then
				self:DebugPrint(("Received result of type %s (instead of table) with value %s."):format(type(res), res))
			else
				if self.db.profile.debugging then
					self:LogFrameAdd(("|cff00ff00%s (%s): |r"):format(sender, pcls))
				end
				-- go through all buffs in the profile (if there is one that is...)
				for k1,v1 in pairs(actbuff.profile) do
					if v1 then
						requiredbuffs[#requiredbuffs + 1] = k1
						self:DebugPrint(("  Required buff: %s -> %s"):format(k1, v1 and YES or NO))
					end
				end
			
				-- get the details of all the buff from the profile 
				for k1,v1 in pairs(requiredbuffs) do
					self:DebugPrint(("%s | %s"):format(k1, v1))
					local user    = actbase.profile[v1].user     --caster of the buff
					local sources = actbase.profile[v1].sources  --equal buffs
					local found   = false

					for k2,v2 in pairs(sources) do -- there might be several equivalent buffs
						-- v2 = buffname(s)
						for k3,v3 in pairs(res) do -- go through the retrieved buffs
							-- v3 = bufftable(s)
							-- normal case (names match)
							self:DebugPrint(("%s | %s | %s | %s | %s"):format(k2, v2[pcls], v2.Weight, k3, v3.name))

							if((v3.name == k2 or v3.name == self:ReverseTranslateSafe(k2)) and not found and (v3.time > secs or v3.time == 0)) then
								found = true
								buffval = buffval + v2[pcls]*v2.Weight
								if self.db.profile.debugging then
									self:LogFrameAdd(("  |cff00ff00%s -> %s | %s | %s (%s) >> %s|r"):format(k2, v3["name"], v3["type"], self:FormatTime(v3["time"]), user, found and YES or NO ))
								end
							end
							-- special case (flask and elixir match only partial)
							if((k2 == "Flask" or k2 == "Battle Elixir" or k2 == "Guardian Elixir") and self:StringContains(v3["ttip"], k2) and not found and (v3.time > secs or v3.time == 0)) then
								found = true
								buffval = buffval + v2[pcls]*v2.Weight
								if self.db.profile.debugging then
									self:LogFrameAdd(("  |cff00ff00%s -> %s | %s | %s (%s) >> %s|r"):format(k2, v3["name"], v3["type"], self:FormatTime(v3["time"]), user, found and YES or NO ))
								end
							end
							
							-- special case (weapon buff has no name)
							if(k2 == "Weapon" and v3["type"] == "mhw" and not found and (v3.time > secs or v3.time == 0)) then
								found = true
								buffval = buffval + v2[pcls]*v2.Weight
								if self.db.profile.debugging then
									self:LogFrameAdd(("  |cff00ff00%s -> %s | %s | %s (%s) >> %s|r"):format(k2, v3["name"], v3["type"], self:FormatTime(v3["time"]), user, found and YES or NO ))
								end
							end
							if (not found and self.db.profile.debugging) then
								self:LogFrameAdd(("  %s -> %s | %s | %s (%s) >> %s"):format(k2, v3["name"], v3["type"], self:FormatTime(v3["time"]), user, found and YES or NO ))
							end
						end	
					end
					if (not found and self.db.profile.debugging) then
						self:LogFrameAdd(("  |cffff0000%s -> missing (job of %s)|r"):format(v1, user))
					end
				end
			end		
		end
	end
	return buffval
end

function fmBuffCheck:Ucfirst(str)
	return string.upper(string.sub(str,1,1)) .. string.lower(string.sub(str,2))
end

-- Frames and their management function...

-- LogFrame

function fmBuffCheck:LogFrameHide()
	fmBuffCheck.LogFrame:Hide()
end

function fmBuffCheck:LogFrameUpdate()
	fmBuffCheck.LogFrame.Scroll:UpdateScrollChildRect()
	fmBuffCheck.LogFrame.Box:ClearFocus()
end

function fmBuffCheck:LogFrameClear()
	fmBuffCheck.LogFrame.Box:SetText("")
end

function fmBuffCheck:LogFrameShow()
	if self.db.profile.showlogframe then
		fmBuffCheck.LogFrame:Show()
	else
		fmBuffCheck.LogFrame:Hide()
	end
end

function fmBuffCheck:LogFrameAdd(txt)
	fmBuffCheck.LogFrame.Box:SetText(self.LogFrame.Box:GetText() .. "\n" .. txt)
end

function fmBuffCheck:CreateLogFrame()
	fmBuffCheck.LogFrame = CreateFrame("Frame", "", UIParent)
	fmBuffCheck.LogFrame:Hide()
	fmBuffCheck.LogFrame:SetPoint("CENTER", "UIParent", "CENTER")
	fmBuffCheck.LogFrame:SetFrameStrata("TOOLTIP")
	fmBuffCheck.LogFrame:SetHeight(280)
	fmBuffCheck.LogFrame:SetWidth(500)
	fmBuffCheck.LogFrame:SetBackdrop({
		bgFile = "Interface/Tooltips/UI-Tooltip-Background",
		edgeFile = "Interface/Tooltips/UI-Tooltip-Border",
		tile = true, tileSize = 32, edgeSize = 32,
		insets = { left = 9, right = 9, top = 9, bottom = 9 }
	})
	fmBuffCheck.LogFrame:SetBackdropColor(0,0,0, 0.8)

	fmBuffCheck.LogFrame.Done = CreateFrame("Button", "", fmBuffCheck.LogFrame, "OptionsButtonTemplate")
	fmBuffCheck.LogFrame.Done:SetText("Close")
	fmBuffCheck.LogFrame.Done:SetPoint("BOTTOMRIGHT", fmBuffCheck.LogFrame, "BOTTOMRIGHT", -10, 10)
	fmBuffCheck.LogFrame.Done:SetScript("OnClick", self.LogFrameHide)

	fmBuffCheck.LogFrame.Mesg = fmBuffCheck.LogFrame:CreateFontString("", "OVERLAY", "GameFontNormalSmall")
	fmBuffCheck.LogFrame.Mesg:SetJustifyH("LEFT")
	fmBuffCheck.LogFrame.Mesg:SetPoint("TOPRIGHT", fmBuffCheck.LogFrame.Done, "TOPLEFT", -10, 0)
	fmBuffCheck.LogFrame.Mesg:SetPoint("LEFT", fmBuffCheck.LogFrame, "LEFT", 15, 0)
	fmBuffCheck.LogFrame.Mesg:SetHeight(20)
	fmBuffCheck.LogFrame.Mesg:SetText(L["You can select all and copy above for further study..."])

	fmBuffCheck.LogFrame.Scroll = CreateFrame("ScrollFrame", "fmBuffCheckLogInputScroll", fmBuffCheck.LogFrame, "UIPanelScrollFrameTemplate")
	fmBuffCheck.LogFrame.Scroll:SetPoint("TOPLEFT", fmBuffCheck.LogFrame, "TOPLEFT", 20, -20)
	fmBuffCheck.LogFrame.Scroll:SetPoint("RIGHT", fmBuffCheck.LogFrame, "RIGHT", -30, 0)
	fmBuffCheck.LogFrame.Scroll:SetPoint("BOTTOM", fmBuffCheck.LogFrame.Done, "TOP", 0, 10)

	fmBuffCheck.LogFrame.Box = CreateFrame("EditBox", "fmBuffCheckLogEditBox", fmBuffCheck.LogFrame.Scroll)
	fmBuffCheck.LogFrame.Box:SetWidth(450)
	fmBuffCheck.LogFrame.Box:SetHeight(85)
	fmBuffCheck.LogFrame.Box:SetMultiLine(true)
	fmBuffCheck.LogFrame.Box:SetAutoFocus(false)
	fmBuffCheck.LogFrame.Box:SetFontObject(GameTooltipTextSmall)
	fmBuffCheck.LogFrame.Box:SetScript("OnEscapePressed", fmBuffCheck.LogFrameHide)
	fmBuffCheck.LogFrame.Box:SetScript("OnTextChanged", fmBuffCheck.LogFrameUpdate)
	fmBuffCheck.LogFrame.Scroll:SetScrollChild(fmBuffCheck.LogFrame.Box)
end

-- ResFrame

function fmBuffCheck:ResFrameClear()
	fmBuffCheck:ResFrameWrite("", "", "")
end

function fmBuffCheck:ResFrameShow()
	if self.db.profile.showresframe and self:IsAuthorized() then
		fmBuffCheckResFrame:Show()
	else
		fmBuffCheckResFrame:Hide()
	end
end

function fmBuffCheck:ResFrameWrite(txt1, txt2, txt3)
	if (txt1 ~= nil) then
		fmBuffCheckResFrameInnerText1:SetText(txt1)
	end
	if (txt2 ~= nil) then
		fmBuffCheckResFrameInnerText2:SetText(txt2)
	end
	if (txt3 ~= nil) then
		fmBuffCheckResFrameInnerText3:SetText(txt3)
	end
end

function fmBuffCheck:ResFrameHide()
	fmBuffCheckResFrame:Hide()
end

function fmBuffCheck:ResFrameTooltipShow()
	local p = self.lastbuffresults.p
	local t = self.lastbuffresults.t

	fmBuffCheckResFrameTooltip:SetOwner(fmBuffCheckResFrame, "ANCHOR_BOTTOMLEFT")
	fmBuffCheckResFrameTooltip:AddLine("fmBuffCheck Report Details:")
	fmBuffCheckResFrameTooltip:AddLine(" ")

	if (p ~= nil) then
		for k1, v1 in pairs(p) do
			fmBuffCheckResFrameTooltip:AddDoubleLine(v1.name, v1.values[1].value * v1.fact, 1, 1, 0, 1, 1, 1)
		end
	else
		fmBuffCheckResFrameTooltip:AddLine(L["No results yet."])
	end
--[[					players[name]["name"]   = name
					players[name]["source"] = self.repbuf[name]
					players[name]["pcls"]   = self:ReverseTranslateSafe(self:Ucfirst(pcls))
					players[name]["fact"]   = classFact[players[name]["pcls"] ]
					players[name]["values"] = {}

		local t =
		{
			{
				t = 10,
				b = 0,
				w = 100,
				a = 0,
				n = "",
				s = "",
			}, --[1]
]]
	--fmBuffCheckResFrameTooltip:AddDoubleLine(L["Request Timeout:"], self.db.profile.requesttimeout .. "s", 1, 1, 0, 1, 1, 1)
	--fmBuffCheckResFrameTooltip:AddLine(" ")
	fmBuffCheckResFrameTooltip:Show()
end

function fmBuffCheck:ResFrameTooltipHide()
	fmBuffCheckResFrameTooltip:Hide()
end

function fmBuffCheck:GetResFrameMenu()
	local k1, v1, fmt
	local p = self.lastbuffresults.p
	local ret =
	{
		type = 'group',
		handler = fmBuffCheck,
		args = {}
	}
			
	if (p ~= nil) then
		for k1, v1 in pairs(p) do
			if v1.source ~= nil then
				fmt = "|cff00ff00%s (%3.0f)|r"
				ret.args[#ret.args + 1] =
				{
					name = fmt:format(v1.name, v1.values[1].value * v1.fact),
					desc = fmt:format(v1.name, v1.values[1].value * v1.fact),
					type = 'group',
					args = fmBuffCheck:GetResFrameMenuSub(v1.name),
				}
			else
				fmt = "|cffff0000%s|r"
				ret.args[#ret.args + 1] =
				{
					name = fmt:format(v1.name),
					desc = fmt:format(v1.name),
					type = 'header',
				}
			end
			
		end
	end
	
	return ret
end

function fmBuffCheck:GetResFrameMenuSub(name)
	local k1, v1
	local p = self.lastbuffresults.p[name].source
	local ret = {}
	
	if (p ~= nil) then
		for k1, v1 in pairs(p) do
			ret[#ret + 1] =
			{
				--["ttip"] = ttip
				name = ("%s (%s)"):format(self:TranslateSafe(v1.name), self:FormatTimeShort(v1.time)),
				desc = ("%s (%s)"):format(self:TranslateSafe(v1.name), self:FormatTimeShort(v1.time)),
				type = 'group',
				args = {{
					name = v1.ttip,
					type = 'header',
				}},
			}
		end
	end
	
	return ret
end